#include <stdio.h>

#include "common.h"
#include "macros.h"
#include "as_tree.h"

typedef data_type_t (*guess_handler_t)(struct ast_node *);

static data_type_t id_data_type(struct ast_node *node);
static data_type_t refer_data_type(struct ast_node *node);
static data_type_t const_data_type(struct ast_node *node);
#ifdef _MEM_AREA
static data_type_t mem_area_data_type(struct ast_node *node);
#endif
static data_type_t call_ret_type(struct ast_node *node);
static data_type_t access_data_type(struct ast_node *node);
static data_type_t arith_op_data_type(struct ast_node *node);
static data_type_t bool_op_data_type(struct ast_node *node);
static data_type_t assignment_data_type(struct ast_node *node);
data_type_t guess_data_type(struct ast_node *node);

/* Sometimes we need guess the data type of the expression, we return the
 * first found data type */
static struct {
	node_type_t node_type;
	guess_handler_t	handler;
} guessers[] = {
	{ NODE_TYPE_ID,		id_data_type },
	{ NODE_TYPE_REFER,	refer_data_type },
	{ NODE_TYPE_CONST,	const_data_type },
#ifdef _MEM_AREA
	{ NODE_TYPE_MEM_AREA,	mem_area_data_type },
#endif
	{ NODE_TYPE_ACCESS,	access_data_type },
	{ NODE_TYPE_CALL,	call_ret_type },
	{ NODE_TYPE_ASSIGN,	assignment_data_type },
	{ NODE_TYPE_SUM_OP,	arith_op_data_type },
	{ NODE_TYPE_MUL_OP,	arith_op_data_type },
	{ NODE_TYPE_REL_OP,	bool_op_data_type },
	{ NODE_TYPE_AND_OP,	bool_op_data_type },
	{ NODE_TYPE_OR_OP,	bool_op_data_type },
	{ NODE_TYPE_NOT,	bool_op_data_type  },
	{ NODE_TYPE_UNKNOWN,	NULL }
};

static data_type_t id_data_type(struct ast_node *node)
{
	struct ast_node_id *id;

	return_val_if_fail(node != NULL, DATA_TYPE_UNKNOWN);

	id = (struct ast_node_id *)node;

	return id->decl_type;
}

static data_type_t refer_data_type(struct ast_node *node)
{
	struct ast_node_refer *refer;

	return_val_if_fail(node != NULL, DATA_TYPE_UNKNOWN);

	refer = (struct ast_node_refer *)node;

	return refer->decl_type;
}

static data_type_t const_data_type(struct ast_node *node)
{
	struct ast_node_const *leaf;

	return_val_if_fail(node != NULL, DATA_TYPE_UNKNOWN);

	leaf = (struct ast_node_const *)node;

	return leaf->data_type;
}

#ifdef _MEM_AREA
static data_type_t mem_area_data_type(struct ast_node *node)
{
	struct ast_node_mem_area *mem_area;

	return_val_if_fail(node != NULL, DATA_TYPE_UNKNOWN);

	mem_area = (struct ast_node_mem_area *)node;

	if (mem_area->nbits != -1)
		return DATA_TYPE_INTEGER;

	switch (mem_area->mem_data_type) {
	case 'B':
	case 'S':
	case 'L':
		return DATA_TYPE_INTEGER;
	case 'F':
		return DATA_TYPE_REAL;
	}

	return DATA_TYPE_UNKNOWN;
}
#endif

static data_type_t call_ret_type(struct ast_node *node)
{
	struct ast_node_func_call *call;

	return_val_if_fail(node != NULL, DATA_TYPE_UNKNOWN);

	call = (struct ast_node_func_call *)node;

	return call->ret_type;
}

static data_type_t access_data_type(struct ast_node *node)
{
	struct ast_node_access *access;

	return_val_if_fail(node != NULL, DATA_TYPE_UNKNOWN);

	access = (struct ast_node_access *)node;

	return access->decl_type;
}

static data_type_t assignment_data_type(struct ast_node *node)
{
	struct ast_node_assign *assign;
	data_type_t type;
	unsigned int i;

	return_val_if_fail(node != NULL, DATA_TYPE_UNKNOWN);

	assign = (struct ast_node_assign *)node;
	type = DATA_TYPE_UNKNOWN;

	for (i = 0; i < assign->nleft; i++) {
		type = guess_data_type(assign->left[i]);
		if (type != DATA_TYPE_UNKNOWN)
			return type;
	}

	type = guess_data_type(assign->right);

	return type;
}

static data_type_t bool_op_data_type(struct ast_node *node)
{
	return_val_if_fail(node != NULL, DATA_TYPE_UNKNOWN);
	return DATA_TYPE_INTEGER;
}

static data_type_t arith_op_data_type(struct ast_node *node)
{
	struct ast_node_op *op;
	data_type_t type;

	return_val_if_fail(node != NULL, DATA_TYPE_UNKNOWN);

	op = (struct ast_node_op *)node;

	type = guess_data_type(op->left);

	if (type != DATA_TYPE_UNKNOWN)
		return type;

	type = guess_data_type(op->right);

	return type;
}

data_type_t guess_data_type(struct ast_node *node)
{
	data_type_t type;
	unsigned int i;

	return_val_if_fail(node != NULL, DATA_TYPE_UNKNOWN);

	type = DATA_TYPE_UNKNOWN;

	for (i = 0; guessers[i].handler != NULL; i++) {
		if (node->type == guessers[i].node_type) {
			type = guessers[i].handler(node);
			break;
		}
	}

	return type;
}
